#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef USE_NLLIBC
#include <nllibc.h>
#endif

#include "lib.h"
#include "objlist.h"
#include "type.h"
#include "syntax.h"
#include "symbol.h"
#include "asm.h"
#include "asm_syntax.h"
#include "asm_symbol.h"

typedef enum {
  SECTION_TYPE_NONE,
  SECTION_TYPE_TEXT,
  SECTION_TYPE_DATA,
} section_type_t;

static section_type_t section_type = SECTION_TYPE_NONE;

static int section_change(FILE *out, section_type_t type)
{
  if (section_type != type) {
    fprintf(out, "\t.section\t");
    switch (type) {
    case SECTION_TYPE_TEXT: fprintf(out, ".text"); break;
    case SECTION_TYPE_DATA: fprintf(out, ".data"); break;
    case SECTION_TYPE_NONE:
      fprintf(out, "unknown");
      break;
    }
    fprintf(out, "\n\n");
  }

  return section_type = type;
}

int asm_section_change_text(FILE *out)
{
  return section_change(out, SECTION_TYPE_TEXT);
}

int asm_section_change_data(FILE *out)
{
  return section_change(out, SECTION_TYPE_DATA);
}

int asm_syntax_list(FILE *out, objlist_t syntax_list)
{
  objentry_t entry;
  syntax_t syntax;

  for (entry = objlist_get_head(syntax_list);
       !objlist_is_end(syntax_list, entry);
       entry = objentry_get_next(entry)) {
    syntax = objentry_get_syntax(entry);
    asm_syntax(out, syntax);
  }

  return 0;
}

int asm_symbol_list(FILE *out, objlist_t symbol_list)
{
  objentry_t entry;
  symbol_t symbol;

  for (entry = objlist_get_head(symbol_list);
       !objlist_is_end(symbol_list, entry);
       entry = objentry_get_next(entry)) {
    symbol = objentry_get_obj(entry);
    asm_symbol(out, symbol);
  }

  return 0;
}
