#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef USE_NLLIBC
#include <nllibc.h>
#endif

#include "lib.h"
#include "objlist.h"
#include "type.h"
#include "syntax.h"
#include "asm.h"
#include "asm_syntax.h"
#include "asm_symbol.h"

static int asm_syntax_symbol(FILE *out, symbol_t symbol)
{
  asm_symbol_value_set(out, symbol);
  if (symbol->flags & SYMBOL_FLAG_LOCATED)
    asm_symbol_value_get(out, symbol);
  return 0;
}

static char *r0_by_size(model_t model)
{
  switch (model_get_size(model)) {
  case -1: ERREXIT(1);
  case  1: return "al";
  case  2: return "ax";
  default: break;
  }
  return "eax";
}

static int asm_syntax_ifgoto(FILE *out, syntax_ifgoto_t ifgoto)
{
  asm_symbol_value_set(out, ifgoto->cond);
  asm_symbol_value_get(out, ifgoto->cond);
  fprintf(out, "\tcmp\t$0, %%%s\n", r0_by_size(ifgoto->cond->model));
  fprintf(out, "\tjne\t%s\n", ifgoto->label->obj.label.name);
  return 0;
}

static int asm_syntax_ifngoto(FILE *out, syntax_ifngoto_t ifngoto)
{
  asm_symbol_value_set(out, ifngoto->cond);
  asm_symbol_value_get(out, ifngoto->cond);
  fprintf(out, "\tcmp\t$0, %%%s\n", r0_by_size(ifngoto->cond->model));
  fprintf(out, "\tje\t%s\n", ifngoto->label->obj.label.name);
  return 0;
}

static int asm_syntax_goto(FILE *out, syntax_goto_t _goto)
{
  fprintf(out, "\tjmp\t%s\n", _goto->label->obj.label.name);
  return 0;
}

static int asm_syntax_label(FILE *out, syntax_label_t label)
{
  fprintf(out, "%s:\n", label->name);
  return 0;
}

static int asm_syntax_return(FILE *out, syntax_return_t _return)
{
  if (_return->value != NULL) {
    asm_symbol_value_set(out, _return->value);
    asm_symbol_value_get(out, _return->value);
  }
  fprintf(out, "\tjmp\t%s\n", _return->function->obj.function.retlabel->obj.label.name);
  return 0;
}

int asm_syntax(FILE *out, syntax_t syntax)
{
  saveline(syntax->line);

  switch (syntax->type) {
  case C_TYPE_BLOCK:
    asm_syntax_list(out, syntax->obj.block.syntaxes);
    break;

  case C_TYPE_SYMBOL:
    asm_syntax_symbol(out, syntax->obj.symbol);
    break;

  case C_TYPE_IFGOTO:
    asm_syntax_ifgoto(out, &syntax->obj.ifgoto);
    break;

  case C_TYPE_IFNGOTO:
    asm_syntax_ifngoto(out, &syntax->obj.ifngoto);
    break;

  case C_TYPE_GOTO:
    asm_syntax_goto(out, &syntax->obj._goto);
    break;

  case C_TYPE_LABEL:
    asm_syntax_label(out, &syntax->obj.label);
    break;

  case C_TYPE_RETURN:
    asm_syntax_return(out, &syntax->obj._return);
    break;

  case C_TYPE_NOP:
    break;

  default:
    ERREXIT(1);
  }

  return 0;
}
