#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <nllibc.h>

struct area_header {
  struct area_header *next;
  struct area_header *prev;
  int size;
  int magic;
#define AREA_HEADER_MAGIC_NUMBER 0x12345678
};

static struct area_header *alloc_start = NULL;
static struct area_header *alloc_end = NULL;

static char *extend_memory(void)
{
  static char *p = NULL;
  if (p != NULL) {
    if (sbrk(0x1000) == (void *)-1)
      return NULL;
  }
  p = sbrk(0);
  if (p == (void *)-1)
    p = NULL;
  return p;
}

static int extend(void)
{
  struct area_header *new_alloc_end;

  new_alloc_end = (struct area_header *)extend_memory();
  if (new_alloc_end == NULL)
    return -1;

  new_alloc_end--;

  alloc_end->prev->next = new_alloc_end;
  new_alloc_end->prev = alloc_end->prev;
  new_alloc_end->next = NULL;
  new_alloc_end->size = 0;
  new_alloc_end->magic = 0;

  alloc_end = new_alloc_end;

  return 0;
}

#if 0
static int expand(int size)
{
  struct area_header *new_alloc_start;
  struct area_header *new_alloc_end;
  int fd = -1;
  char *p;

  fd = open("/dev/zero", O_RDWR);
  if (fd < 0)
    goto err;

  p = (char *)mmap(NULL, size, PROT_READ|PROT_WRITE, MAP_PRIVATE, fd, 0);
  if (p == NULL)
    goto err;

  new_alloc_start = (struct area_header *)p;
  new_alloc_end = (struct area_header *)(p + size);

  new_alloc_end--;

  alloc_end->next = new_alloc_start;
  alloc_end->size = -1; /* for skip */

  new_alloc_start->prev = alloc_end;
  new_alloc_start->next = new_alloc_end;
  new_alloc_end->prev = new_alloc_start;
  new_alloc_end->next = NULL;
  new_alloc_start->size = 0;
  new_alloc_end->size = 0;
  new_alloc_start->magic = 0;
  new_alloc_end->magic = 0;

  alloc_end = new_alloc_end;

  return 0;

err:
  if (fd >= 0)
    close(fd);

  return -1;
}
#endif

static int init(void)
{
  if (alloc_start == NULL) {
    alloc_start = (struct area_header *)extend_memory();
    alloc_end = (struct area_header *)extend_memory();

    if ((alloc_start == NULL) || (alloc_end == NULL))
      return -1;

    alloc_end--;

    alloc_start->prev = NULL;
    alloc_start->next = alloc_end;
    alloc_end->prev = alloc_start;
    alloc_end->next = NULL;
    alloc_start->size = 0;
    alloc_end->size = 0;
    alloc_start->magic = 0;
    alloc_end->magic = 0;
  }
  return 0;
}

void *malloc(size_t size)
{
  struct area_header *area, *following;
  int aligned_size;

  if (init() < 0)
    return NULL;

  aligned_size = ((size + 3) / 4) * 4;

  area = alloc_start;
  while (1) {
    for (; area->next; area = area->next) {
      if (area->size < 0) /* skip */
	continue;
      following = (struct area_header *)((char *)(area + 1) + area->size);
      if ((char *)area->next >= (char *)(following + 1) + aligned_size) {
	following->size = aligned_size;
	following->magic = AREA_HEADER_MAGIC_NUMBER;
	following->prev = area;
	following->next = area->next;
	area->next->prev = following;
	area->next = following;
	return following + 1;
      }
    }

    area = area->prev;

    if (extend() < 0)
      break;
  }

  return NULL;
}

void *calloc(size_t number, size_t size)
{
  int s = size * number;
  void *p;

  p = malloc(s);
  if (p != NULL) {
    memset(p, 0, s);
  }

  return p;
}

void *realloc(void *ptr, size_t size)
{
  struct area_header *area;
  int aligned_size;
  void *p;

  if (init() < 0)
    return NULL;

  if (ptr != NULL) {
    aligned_size = ((size + 3) / 4) * 4;

    area = ((struct area_header *)ptr) - 1;
    if (aligned_size <= area->size)
      return ptr;

    if ((char *)area->next >= (char *)(area + 1) + aligned_size) {
      area->size = aligned_size;
      return ptr;
    }
  }

  p = malloc(size);
  if (p == NULL)
    return NULL;

  if (ptr) {
    memcpy(p, ptr, size);
    free(ptr);
  }

  return p;
}

void free(void *ptr)
{
  struct area_header *area;

  if (ptr == NULL)
    return;

  area = ((struct area_header *)ptr) - 1;

  if (area->magic != AREA_HEADER_MAGIC_NUMBER) {
    ;
  }

  area->prev->next = area->next;
  area->next->prev = area->prev;
  area->prev = NULL;
  area->next = NULL;
  area->size = 0;
  area->magic = 0;

  return;
}
